import functools
from collections import namedtuple
from unittest.mock import create_autospec

import pika
import pytest

from yapw.methods.blocking import ack, nack, publish
from yapw.util import default_encode

Connection = namedtuple("Connection", "is_open add_callback_threadsafe")
Channel = namedtuple("Channel", "channel_number is_open basic_ack basic_nack basic_publish")
State = namedtuple("State", "format_routing_key connection exchange encode content_type delivery_mode")

ack_nack_parameters = [(ack, "ack", [1]), (nack, "nack", [1])]
parameters = ack_nack_parameters + [(publish, "publish", [{"message": "value"}, "q"])]


def format_routing_key(exchange, routing_key):
    return f"{exchange}_{routing_key}"


def dumps(message, content_type):
    return b"overridden"


@pytest.mark.parametrize(
    "encode,content_type,body",
    [(default_encode, "application/json", b'{"message":"value"}'), (dumps, "text/plain", b"overridden")],
)
def test_publish(encode, content_type, body):
    connection = create_autospec(Connection, is_open=True)
    channel = create_autospec(Channel, channel_number=1, is_open=True)
    function = functools.partial(format_routing_key, "exch")
    state = create_autospec(
        State,
        connection=connection,
        format_routing_key=function,
        exchange="exch",
        encode=encode,
        content_type=content_type,
        delivery_mode=2,
    )

    publish(state, channel, {"message": "value"}, "q")

    connection.add_callback_threadsafe.assert_called_once()

    cb = connection.add_callback_threadsafe.call_args[0][0]
    cb()

    properties = pika.BasicProperties(delivery_mode=2, content_type=content_type)
    getattr(channel, "basic_publish").assert_called_once_with(
        exchange="exch", routing_key="exch_q", body=body, properties=properties
    )


@pytest.mark.parametrize("function,infix,args", ack_nack_parameters)
@pytest.mark.parametrize("kwargs", [{}, {"multiple": True}])
def test_ack_nack(function, infix, args, kwargs):
    connection = create_autospec(Connection, is_open=True)
    channel = create_autospec(Channel, channel_number=1, is_open=True)
    state = create_autospec(State, connection=connection)

    function(state, channel, *args, **kwargs)

    connection.add_callback_threadsafe.assert_called_once()

    cb = connection.add_callback_threadsafe.call_args[0][0]
    cb()

    getattr(channel, f"basic_{infix}").assert_called_once_with(*args, **kwargs)


@pytest.mark.parametrize("function,infix,args", parameters)
def test_channel_closed(function, infix, args, caplog):
    connection = create_autospec(Connection, is_open=True)
    channel = create_autospec(Channel, channel_number=1, is_open=False)
    state = create_autospec(State, connection=connection)

    function(state, channel, *args)

    connection.add_callback_threadsafe.assert_called_once()

    cb = connection.add_callback_threadsafe.call_args[0][0]
    cb()

    getattr(channel, f"basic_{infix.lower()}").assert_not_called()

    assert len(caplog.records) == 1
    assert caplog.records[-1].levelname == "ERROR"
    assert caplog.records[-1].message == f"Can't {infix} as channel is closed or closing"


@pytest.mark.parametrize("function,infix,args", parameters)
def test_connection_closed(function, infix, args, caplog):
    connection = create_autospec(Connection, is_open=False)
    channel = create_autospec(Channel, channel_number=1, is_open=True)
    state = create_autospec(State, connection=connection)

    function(state, channel, *args)

    connection.add_callback_threadsafe.assert_not_called()

    assert len(caplog.records) == 1
    assert caplog.records[-1].levelname == "ERROR"
    assert caplog.records[-1].message == f"Can't {infix} as connection is closed or closing"



# Copied and adapted from https://github.com/scrapy/scrapy/blob/master/scrapy/utils/ossignal.py
import signal
from types import FrameType
from typing import Callable, Optional, Union

signal_names = {}
for signame in dir(signal):
    if signame.startswith("SIG") and not signame.startswith("SIG_"):
        signum = getattr(signal, signame)
        if isinstance(signum, int):
            signal_names[signum] = signame


def install_signal_handlers(function: Union[Callable[[int, Optional[FrameType]], None], signal.Handlers]) -> None:
    """
    Installs handlers for the SIGTERM, SIGINT, SIGUSR1 and SIGUSR2 signals.

    :param function: the handler
    """
    signal.signal(signal.SIGTERM, function)
    signal.signal(signal.SIGINT, function)
    signal.signal(signal.SIGUSR1, function)
    signal.signal(signal.SIGUSR2, function)



import pytest

from yapw.clients import Base


def test_format_routing_key_default():
    client = Base()

    formatted = client.format_routing_key("test")

    assert formatted == "test"


def test_format_routing_key_kwargs():
    client = Base(routing_key_template="{exchange}_{routing_key}")
    client.exchange = "prefix"

    formatted = client.format_routing_key("test")

    assert formatted == "prefix_test"


def test_format_routing_key_invalid():
    client = Base(routing_key_template="{invalid}_{routing_key}")

    with pytest.raises(KeyError):
        client.format_routing_key("test")



#!/usr/bin/env python
"""
This script can be used to test error handling.
"""

import logging

from yapw import clients
from yapw.decorators import requeue

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


class Client(clients.Threaded, clients.Transient, clients.Blocking, clients.Base):
    pass


def callback(state, channel, method, properties, body):
    raise Exception("message")


def main():
    client = Client(exchange="yapw_development")
    client.consume(callback, "raise", decorator=requeue)


if __name__ == "__main__":
    main()



import logging
from unittest.mock import call, patch

import pika
import pytest

from yapw.clients import Base, Blocking, Durable, Transient
from yapw.util import default_encode


class DurableClient(Durable, Blocking, Base):
    pass


class TransientClient(Transient, Blocking, Base):
    pass


def dumps(message, content_type):
    return b"overridden"


@pytest.mark.parametrize("client_class", [DurableClient, TransientClient])
@patch("pika.BlockingConnection")
def test_init_default(connection, client_class):
    client = client_class()

    client.channel.exchange_declare.assert_not_called()

    assert client.exchange == ""
    assert client.encode == default_encode
    assert client.content_type == "application/json"
    assert client.format_routing_key("test") == "_test"


@pytest.mark.parametrize("client_class,durable", [(DurableClient, True), (TransientClient, False)])
@patch("pika.BlockingConnection")
def test_init_kwargs(connection, client_class, durable):
    client = client_class(
        exchange="exch",
        exchange_type="fanout",
        encode=dumps,
        content_type="application/octet-stream",
        routing_key_template="{routing_key}_{exchange}",
    )

    client.channel.exchange_declare.assert_called_once_with(exchange="exch", exchange_type="fanout", durable=durable)

    assert client.exchange == "exch"
    assert client.encode == dumps
    assert client.content_type == "application/octet-stream"
    assert client.format_routing_key("test") == "test_exch"


@pytest.mark.parametrize("client_class,durable", [(DurableClient, True), (TransientClient, False)])
@patch("pika.BlockingConnection")
def test_declare_queue(connection, client_class, durable):
    client = client_class(exchange="exch")

    client.declare_queue("q")

    client.channel.queue_declare.assert_called_once_with(queue="exch_q", durable=durable, arguments=None)
    assert client.channel.queue_bind.call_count == 1
    client.channel.queue_bind.assert_has_calls(
        [
            call(exchange="exch", queue="exch_q", routing_key="exch_q"),
        ]
    )


@pytest.mark.parametrize("client_class,durable", [(DurableClient, True), (TransientClient, False)])
@patch("pika.BlockingConnection")
def test_declare_queue_routing_keys(connection, client_class, durable):
    client = client_class(exchange="exch")

    client.declare_queue("q", ["r", "k"])

    client.channel.queue_declare.assert_called_once_with(queue="exch_q", durable=durable, arguments=None)
    assert client.channel.queue_bind.call_count == 2
    client.channel.queue_bind.assert_has_calls(
        [
            call(exchange="exch", queue="exch_q", routing_key="exch_r"),
            call(exchange="exch", queue="exch_q", routing_key="exch_k"),
        ]
    )


@pytest.mark.parametrize("client_class,delivery_mode", [(DurableClient, 2), (TransientClient, 1)])
@patch("pika.BlockingConnection")
def test_publish(connection, client_class, delivery_mode, caplog):
    connection.return_value.channel.return_value.channel_number = 1

    caplog.set_level(logging.DEBUG)

    client = client_class(exchange="exch")

    client.publish({"a": 1}, "q")

    properties = pika.BasicProperties(delivery_mode=delivery_mode, content_type="application/json")
    client.channel.basic_publish.assert_called_once_with(
        exchange="exch", routing_key="exch_q", body=b'{"a":1}', properties=properties
    )

    assert len(caplog.records) == 1
    record = caplog.records[-1]
    assert record.levelname == "DEBUG"
    assert record.message == "Published message {'a': 1} on channel 1 to exchange exch with routing key exch_q"



import sys
from typing import Any, Callable, NamedTuple, Optional

import pika

if sys.version_info[:2] >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict  # https://github.com/python/mypy/issues/1153


#:
Encode = Callable[[Any, str], bytes]
#:
Decode = Callable[[bytes, Optional[str]], Any]


class State(NamedTuple):
    """
    Attributes that can be used safely in consumer callbacks.
    """

    #: A function to format the routing key.
    format_routing_key: Callable[[str], str]
    #: The connection.
    connection: pika.BlockingConnection
    #: The exchange name.
    exchange: str
    #: The message body's encoder.
    encode: Encode
    #: The message's content type.
    content_type: str
    #: The message's delivery mode.
    delivery_mode: int


class PublishKeywords(TypedDict, total=False):
    """
    Keyword arguments for ``basic_publish``.
    """

    #: The exchange to publish to.
    exchange: str
    #: The message's routing key.
    routing_key: str
    #: The message's body.
    body: bytes
    #: The message's content type and delivery mode.
    properties: pika.BasicProperties


#:
ConsumerCallback = Callable[[State, pika.channel.Channel, pika.spec.Basic.Deliver, pika.BasicProperties, Any], None]
#:
Decorator = Callable[
    [Decode, ConsumerCallback, State, pika.channel.Channel, pika.spec.Basic.Deliver, pika.BasicProperties, bytes], None
]



"""
Mixins that can be combined to create a RabbitMQ client. For example:

.. code-block:: python

   from yapw import clients

   class Client(clients.Threaded, clients.Durable, clients.Blocking, clients.Base):
       pass

The layers are:

Base
  For common logic, without interacting with RabbitMQ.

  Available mixins:

  -  :class:`~yapw.clients.Base`
Connection
  Establish a connection to RabbitMQ and create a channel.

  Available mixins:

  -  :class:`~yapw.clients.Blocking`
Publisher
  Declare an exchange, declare and bind queues, and publish messages.

  Available mixins:

  -  :class:`~yapw.clients.Durable`
  -  :class:`~yapw.clients.Transient`
Consumer
  Consume messages.

  Available mixins:

  -  :class:`~yapw.clients.Threaded`

.. note::

   Importing this module sets the level of the "pika" logger to ``WARNING``, so that consumers can use the ``DEBUG``
   and ``INFO`` levels without their messages getting lost in Pika's verbosity.
"""
from __future__ import annotations

import functools
import logging
import signal
from collections import namedtuple
from concurrent.futures import ThreadPoolExecutor
from types import FrameType
from typing import Any, Callable, Dict, List, Optional, Set, Tuple

import pika
import pika.exceptions
from pika.exchange_type import ExchangeType

from yapw.decorators import default_decode, halt
from yapw.ossignal import install_signal_handlers, signal_names
from yapw.types import ConsumerCallback, Decode, Decorator, Encode, State
from yapw.util import basic_publish_debug_args, basic_publish_kwargs, default_encode

logger = logging.getLogger(__name__)

# Pika is verbose.
logging.getLogger("pika").setLevel(logging.WARNING)


def _on_message(
    channel: pika.channel.Channel,
    method: pika.spec.Basic.Deliver,
    properties: pika.BasicProperties,
    body: bytes,
    args: Tuple[ThreadPoolExecutor, Decorator, Decode, ConsumerCallback, State],
) -> None:
    (executor, decorator, decode, callback, state) = args
    executor.submit(decorator, decode, callback, state, channel, method, properties, body)


class Base:
    """
    Provides :meth:`~Base.format_routing_key`, which is used by all methods in other mixins that accept routing keys,
    in order to namespace the routing keys.

    Other mixins should list attributes that can - and are expected to - be used safely in consumer callbacks in a
    ``__safe__`` class attribute.
    """

    __safe__ = ["format_routing_key"]

    def __init__(self, *, routing_key_template: str = "{routing_key}", **kwargs: Any):
        """
        :param routing_key_template:
            a `format string <https://docs.python.org/3/library/string.html#format-string-syntax>`__ that must contain
            the ``{routing_key}`` replacement field and that may contain other fields matching writable attributes
        """
        #: The format string for the routing key.
        self.routing_key_template = routing_key_template  # type: str

    def format_routing_key(self, routing_key: str) -> str:
        """
        Format the routing key.

        :param routing_key: the routing key
        :returns: the formatted routing key
        """
        return self.routing_key_template.format(routing_key=routing_key, **self.__dict__)

    @property  # type: ignore # https://github.com/python/mypy/issues/1362
    @functools.lru_cache(maxsize=None)
    def __getsafe__(self) -> Set[str]:
        """
        Attributes that can be used safely in consumer callbacks, across all base classes.
        """
        return {attr for base in type(self).__bases__ for attr in getattr(base, "__safe__", [])} | set(
            type(self).__safe__
        )


class Blocking:
    """
    Uses a blocking connection while avoiding deadlocks due to
    `blocked connections <https://www.rabbitmq.com/connection-blocked.html>`__.
    """

    # The connection isn't "safe to use" but it can be "used safely" like in:
    # https://github.com/pika/pika/blob/master/examples/basic_consumer_threaded.py
    __safe__ = ["connection"]

    def __init__(
        self,
        *,
        url: str = "amqp://127.0.0.1",
        blocked_connection_timeout: float = 1800,
        prefetch_count: int = 1,
        **kwargs: Any,
    ):
        """
        Connect to RabbitMQ and create a channel.

        :param url: the connection string (don't set a blocked_connection_timeout query string parameter)
        :param blocked_connection_timeout: the timeout, in seconds, that the connection may remain blocked
        :param prefetch_count: the maximum number of unacknowledged deliveries that are permitted on the channel
        """
        super().__init__(**kwargs)

        parameters = pika.URLParameters(url)
        parameters.blocked_connection_timeout = blocked_connection_timeout

        #: The connection.
        self.connection = pika.BlockingConnection(parameters)  # type: pika.BlockingConnection

        #: The channel.
        self.channel = self.connection.channel()  # type: pika.adapters.blocking_connection.BlockingChannel
        self.channel.basic_qos(prefetch_count=prefetch_count)

    def close(self) -> None:
        """
        Close the connection.
        """
        self.connection.close()


class Publisher:
    """
    An abstract parent class. Use :class:`~yapw.clients.Durable` or :class:`~yapw.clients.Transient` instead.
    """

    durable: bool
    delivery_mode: int

    # Attributes that this mixin expects from base classes.
    format_routing_key: Callable[["Publisher", str], str]
    channel: pika.channel.Channel

    __safe__ = ["exchange", "encode", "content_type", "delivery_mode"]

    def __init__(
        self,
        *,
        exchange: str = "",
        exchange_type: ExchangeType = ExchangeType.direct,
        encode: Encode = default_encode,
        content_type: str = "application/json",
        routing_key_template: str = "{exchange}_{routing_key}",
        **kwargs: Any,
    ):
        """
        Declare an exchange, unless using the default exchange.

        When publishing a message, by default, its body is encoded using :func:`yapw.util.default_encode`, and its
        content type is set to "application/json". Use the ``encode`` and ``content_type`` keyword arguments to change
        this. The ``encode`` must be a function that accepts ``(message, content_type)`` arguments and returns bytes.

        :param exchange: the exchange name
        :param exchange_type: the exchange type
        :param encode: the message body's encoder
        :param content_type: the message's content type
        :param routing_key_template: see :meth:`~yapw.clients.Base.__init__`
        """
        super().__init__(routing_key_template=routing_key_template, **kwargs)  # type: ignore # python/mypy#5887

        #: The exchange name.
        self.exchange = exchange  # type: str
        #: The message body's encoder.
        self.encode = encode  # type: Encode
        #: The message's content type.
        self.content_type = content_type  # type: str

        if self.exchange:
            self.channel.exchange_declare(exchange=self.exchange, exchange_type=exchange_type, durable=self.durable)

    def declare_queue(
        self, queue: str, routing_keys: Optional[List[str]] = None, arguments: Optional[Dict[str, str]] = None
    ) -> None:
        """
        Declare a queue, and bind it to the exchange with the routing keys. If no routing keys are provided, the queue
        is bound to the exchange using its name as the routing key.

        :param queue: the queue's name
        :param routing_keys: the queue's routing keys
        """
        if not routing_keys:
            routing_keys = [queue]

        formatted = self.format_routing_key(queue)
        self.channel.queue_declare(queue=formatted, durable=self.durable, arguments=arguments)

        for routing_key in routing_keys:
            routing_key = self.format_routing_key(routing_key)
            self.channel.queue_bind(exchange=self.exchange, queue=formatted, routing_key=routing_key)

    def publish(self, message: Any, routing_key: str) -> None:
        """
        Publish from the main thread, with the provided message and routing key, and with the configured exchange.

        :param message: a decoded message
        :param routing_key: the routing key
        """
        keywords = basic_publish_kwargs(self, message, routing_key)

        self.channel.basic_publish(**keywords)
        logger.debug(*basic_publish_debug_args(self.channel, message, keywords))


class Transient(Publisher):
    """
    Declares a transient exchange, declares transient queues, and uses transient messages.
    """

    durable = False
    delivery_mode = 1


class Durable(Publisher):
    """
    Declares a durable exchange, declares durable queues, and uses persistent messages.
    """

    durable = True
    delivery_mode = 2


# https://github.com/pika/pika/blob/master/examples/basic_consumer_threaded.py
class Threaded:
    """
    Runs the consumer callback in separate threads.
    """

    # Attributes that this mixin expects from base classes.
    format_routing_key: Callable[["Threaded", str], str]
    declare_queue: Callable[["Threaded", str, Optional[List[str]], Optional[Dict[str, str]]], None]
    connection: pika.BlockingConnection
    channel: pika.adapters.blocking_connection.BlockingChannel

    def __init__(self, decode: Decode = default_decode, **kwargs: Any):
        """
        Install signal handlers to stop consuming messages, wait for threads to terminate, and close the connection.

        When consuming a message, by default, its body is decoded using :func:`yapw.decorators.default_decode`. Use the
        ``decode`` keyword argument to change this. The ``decode`` must be a function that accepts ``(state, channel,
        method, properties, body)`` arguments (like the consumer callback) and returns a decoded message.

        :param decode: the message body's decoder
        """
        super().__init__(**kwargs)

        #: The message body's decoder.
        self.decode = decode  # type: Decode

        install_signal_handlers(self._on_shutdown)

    def consume(
        self,
        callback: ConsumerCallback,
        queue: str,
        routing_keys: Optional[List[str]] = None,
        decorator: Decorator = halt,
        arguments: Optional[Dict[str, str]] = None,
    ) -> None:
        """
        Declare a queue, bind it to the exchange with the routing keys, and start consuming messages from that queue.
        If no routing keys are provided, the queue is bound to the exchange using its name as the routing key.

        The consumer callback must be a function that accepts ``(state, channel, method, properties, body)`` arguments,
        all but the first of which are the same as Pika's ``basic_consume``. The ``state`` argument is needed to pass
        attributes to :mod:`yapw.methods.blocking` functions.

        :param callback: the consumer callback
        :param queue: the queue's name
        :param routing_keys: the queue's routing keys
        :param decorator: the decorator of the consumer callback
        :param arguments: the ``arguments`` parameter to the ``queue_declare`` method
        """
        self.declare_queue(queue, routing_keys, arguments)
        formatted = self.format_routing_key(queue)

        # Don't pass `self` to the callback, to prevent use of unsafe attributes and mutation of safe attributes.
        klass = namedtuple("State", self.__getsafe__)  # type: ignore # https://github.com/python/mypy/issues/848
        state = klass(**{attr: getattr(self, attr) for attr in self.__getsafe__})  # type: ignore

        executor = ThreadPoolExecutor(thread_name_prefix=f"yapw-{queue}")
        on_message_callback = functools.partial(_on_message, args=(executor, decorator, self.decode, callback, state))
        self.channel.basic_consume(formatted, on_message_callback)

        logger.debug("Consuming messages on channel %s from queue %s", self.channel.channel_number, formatted)
        try:
            self.channel.start_consuming()
        except pika.exceptions.ConnectionClosedByBroker:
            executor.shutdown()
            # The connection is already closed.
        else:
            executor.shutdown()
            self.connection.close()

    def _on_shutdown(self, signum: int, frame: Optional[FrameType]) -> None:
        install_signal_handlers(signal.SIG_IGN)
        logger.info("Received %s, shutting down gracefully", signal_names[signum])
        self.channel.stop_consuming()



import functools
import json
import logging
import os
import signal
import time

import pytest

from yapw.clients import Base, Blocking, Threaded, Transient
from yapw.decorators import discard, requeue
from yapw.methods.blocking import ack, nack, publish

logger = logging.getLogger(__name__)

DELAY = 0.05
RABBIT_URL = os.getenv("TEST_RABBIT_URL", "amqp://127.0.0.1")


class Client(Threaded, Transient, Blocking, Base):
    pass


def get_client(**kwargs):
    return Client(url=RABBIT_URL, exchange="yapw_test", **kwargs)


def encode(message):
    if not isinstance(message, bytes):
        return json.dumps(message, separators=(",", ":")).encode()
    return message


def kill(signum):
    os.kill(os.getpid(), signum)
    # The signal should be handled once.
    os.kill(os.getpid(), signum)


@pytest.fixture(params=[({}, {"message": "value"}), ({"content_type": "application/octet-stream"}, b"message value")])
def message(request):
    kwargs, body = request.param

    publisher = get_client(**kwargs)
    publisher.declare_queue("q")
    publisher.publish(body, "q")
    yield body
    # Purge the queue, instead of waiting for a restart.
    publisher.channel.queue_purge("yapw_test_q")
    publisher.close()


@pytest.fixture
def short_message(request):
    body = 1

    publisher = get_client()
    publisher.declare_queue("q")
    publisher.publish(body, "q")
    yield body
    # Purge the queue, instead of waiting for a restart.
    publisher.channel.queue_purge("yapw_test_q")
    publisher.close()


# Consumer callbacks.
def sleeper(state, channel, method, properties, body):
    logger.info("Sleep")
    time.sleep(DELAY * 2)
    logger.info("Wake!")
    ack(state, channel, method.delivery_tag)


def raiser(state, channel, method, properties, body):
    raise Exception("message")


def ack_warner(state, channel, method, properties, body):
    logger.warning(body)
    ack(state, channel, method.delivery_tag)


def nack_warner(state, channel, method, properties, body):
    logger.warning(body)
    nack(state, channel, method.delivery_tag)


def writer(state, channel, method, properties, body):
    publish(state, channel, {"message": "value"}, "n")
    ack(state, channel, method.delivery_tag)


# Decoders
def decode(index, body, content_type):
    return body.decode()[index]


@pytest.mark.parametrize(
    "signum,signame",
    [(signal.SIGINT, "SIGINT"), (signal.SIGTERM, "SIGTERM"), (signal.SIGUSR1, "SIGUSR1"), (signal.SIGUSR2, "SIGUSR2")],
)
def test_shutdown(signum, signame, message, caplog):
    caplog.set_level(logging.INFO)

    consumer = get_client()
    consumer.connection.call_later(DELAY, functools.partial(kill, signum))
    consumer.consume(sleeper, "q")

    assert consumer.channel.is_closed
    assert consumer.connection.is_closed

    assert len(caplog.records) == 3
    assert [(r.levelname, r.message) for r in caplog.records] == [
        ("INFO", "Sleep"),
        ("INFO", f"Received {signame}, shutting down gracefully"),
        ("INFO", "Wake!"),
    ]


def test_decode_valid(short_message, caplog):
    consumer = get_client(decode=functools.partial(decode, 0))
    consumer.connection.call_later(DELAY, functools.partial(kill, signal.SIGINT))
    consumer.consume(ack_warner, "q")

    assert consumer.channel.is_closed
    assert consumer.connection.is_closed

    assert len(caplog.records) == 1
    assert caplog.records[-1].levelname == "WARNING"
    assert caplog.records[-1].message == "1"


def test_decode_invalid(short_message, caplog):
    caplog.set_level(logging.INFO)

    consumer = get_client(decode=functools.partial(decode, 10))
    consumer.connection.call_later(DELAY, functools.partial(kill, signal.SIGINT))
    consumer.consume(ack_warner, "q")

    assert consumer.channel.is_closed
    assert consumer.connection.is_closed

    assert len(caplog.records) == 2
    assert [(r.levelname, r.message, r.exc_info is None) for r in caplog.records] == [
        ("ERROR", f"{encode(short_message)} can't be decoded, sending SIGUSR2", False),
        ("INFO", "Received SIGUSR2, shutting down gracefully", True),
    ]


def test_decode_raiser(message, caplog):
    caplog.set_level(logging.INFO)

    consumer = get_client(decode=raiser)
    consumer.connection.call_later(DELAY, functools.partial(kill, signal.SIGINT))
    consumer.consume(ack_warner, "q")

    assert consumer.channel.is_closed
    assert consumer.connection.is_closed

    assert len(caplog.records) == 2
    assert [(r.levelname, r.message, r.exc_info is None) for r in caplog.records] == [
        ("ERROR", f"{encode(message)} can't be decoded, sending SIGUSR2", False),
        ("INFO", "Received SIGUSR2, shutting down gracefully", True),
    ]


def test_halt(message, caplog):
    caplog.set_level(logging.INFO)

    consumer = get_client()
    consumer.connection.call_later(30, functools.partial(kill, signal.SIGINT))  # in case not halted
    consumer.consume(raiser, "q")

    assert consumer.channel.is_closed
    assert consumer.connection.is_closed

    assert len(caplog.records) == 2
    assert [(r.levelname, r.message, r.exc_info is None) for r in caplog.records] == [
        ("ERROR", f"Unhandled exception when consuming {encode(message)}, sending SIGUSR1", False),
        ("INFO", "Received SIGUSR1, shutting down gracefully", True),
    ]


def test_discard(message, caplog):
    caplog.set_level(logging.INFO)

    consumer = get_client()
    consumer.connection.call_later(DELAY, functools.partial(kill, signal.SIGINT))
    consumer.consume(raiser, "q", decorator=discard)

    assert consumer.channel.is_closed
    assert consumer.connection.is_closed

    assert len(caplog.records) == 2
    assert [(r.levelname, r.message, r.exc_info is None) for r in caplog.records] == [
        ("ERROR", f"Unhandled exception when consuming {encode(message)}, discarding message", False),
        ("INFO", "Received SIGINT, shutting down gracefully", True),
    ]


def test_requeue(message, caplog):
    caplog.set_level(logging.INFO)

    consumer = get_client()
    consumer.connection.call_later(DELAY, functools.partial(kill, signal.SIGINT))
    consumer.consume(raiser, "q", decorator=requeue)

    assert consumer.channel.is_closed
    assert consumer.connection.is_closed

    assert len(caplog.records) == 3
    assert [(r.levelname, r.message, r.exc_info is None) for r in caplog.records] == [
        ("ERROR", f"Unhandled exception when consuming {encode(message)} (requeue=True)", False),
        ("ERROR", f"Unhandled exception when consuming {encode(message)} (requeue=False)", False),
        ("INFO", "Received SIGINT, shutting down gracefully", True),
    ]


def test_publish(message, caplog):
    caplog.set_level(logging.DEBUG)

    consumer = get_client()
    consumer.connection.call_later(DELAY, functools.partial(kill, signal.SIGINT))
    consumer.consume(writer, "q")

    assert consumer.channel.is_closed
    assert consumer.connection.is_closed

    assert len(caplog.records) == 5
    assert [(r.levelname, r.message) for r in caplog.records] == [
        ("DEBUG", "Consuming messages on channel 1 from queue yapw_test_q"),
        ("DEBUG", f"Received message {encode(message)} with routing key yapw_test_q and delivery tag 1"),
        (
            "DEBUG",
            "Published message {'message': 'value'} on channel 1 to exchange yapw_test with routing key yapw_test_n",
        ),
        ("DEBUG", "Ack'd message on channel 1 with delivery tag 1"),
        ("INFO", "Received SIGINT, shutting down gracefully"),
    ]


def test_consume_declares_queue(caplog):
    declarer = get_client()
    declarer.connection.call_later(DELAY, functools.partial(kill, signal.SIGINT))
    declarer.consume(raiser, "q")

    publisher = get_client()
    publisher.publish({"message": "value"}, "q")

    consumer = get_client()
    consumer.connection.call_later(DELAY, functools.partial(kill, signal.SIGINT))
    consumer.consume(nack_warner, "q")

    publisher.channel.queue_purge("yapw_test_q")
    publisher.close()

    assert consumer.channel.is_closed
    assert consumer.connection.is_closed

    assert len(caplog.records) > 1
    assert all(r.levelname == "WARNING" and r.message == "{'message': 'value'}" for r in caplog.records)


def test_consume_declares_queue_routing_keys(caplog):
    declarer = get_client()
    declarer.connection.call_later(DELAY, functools.partial(kill, signal.SIGINT))
    declarer.consume(raiser, "q", ["r", "k"])

    publisher = get_client()
    publisher.publish({"message": "r"}, "r")
    publisher.publish({"message": "k"}, "k")

    consumer = get_client()
    consumer.connection.call_later(DELAY, functools.partial(kill, signal.SIGINT))
    consumer.consume(ack_warner, "q", ["r", "k"])

    publisher.channel.queue_purge("yapw_test_q")
    publisher.close()

    assert consumer.channel.is_closed
    assert consumer.connection.is_closed

    assert len(caplog.records) == 2
    assert [(r.levelname, r.message) for r in caplog.records] == [
        ("WARNING", "{'message': 'r'}"),
        ("WARNING", "{'message': 'k'}"),
    ]



"""
Functions for calling channel methods from the context of a consumer callback.
"""

import functools
import logging
from typing import Any, Optional

import pika

from yapw.types import State
from yapw.util import basic_publish_debug_args, basic_publish_kwargs

logger = logging.getLogger(__name__)


def publish(
    state: State, channel: pika.channel.Channel, message: Any, routing_key: str, *args: Any, **kwargs: Any
) -> None:
    """
    Publish with the provided message and routing key, and with the exchange set by the provided state.

    :param state: an object with a ``connection`` attribute
    :param channel: the channel from which to call ``basic_publish``
    :param message: a decoded message
    :param routing_key: the routing key
    """
    keywords = basic_publish_kwargs(state, message, routing_key)
    keywords.update(kwargs)  # type: ignore # https://github.com/python/mypy/issues/4441

    _channel_method_from_thread(state.connection, channel, "publish", *args, **keywords)
    logger.debug(*basic_publish_debug_args(channel, message, keywords))


def ack(state: State, channel: pika.channel.Channel, delivery_tag: Optional[int] = 0, **kwargs: bool) -> None:
    """
    Ack a message by its delivery tag.

    :param state: an object with a ``connection`` attribute
    :param channel: the channel from which to call ``basic_ack``
    :param delivery_tag: the delivery tag
    """
    _channel_method_from_thread(state.connection, channel, "ack", delivery_tag, **kwargs)
    logger.debug("Ack'd message on channel %s with delivery tag %s", channel.channel_number, delivery_tag)


def nack(state: State, channel: pika.channel.Channel, delivery_tag: Optional[int] = 0, **kwargs: bool) -> None:
    """
    Nack a message by its delivery tag.

    :param state: an object with a ``connection`` attribute
    :param channel: the channel from which to call ``basic_nack``
    :param delivery_tag: the delivery tag
    """
    _channel_method_from_thread(state.connection, channel, "nack", delivery_tag, **kwargs)
    logger.debug("Nack'd message on channel %s with delivery tag %s", channel.channel_number, delivery_tag)


def _channel_method_from_thread(
    connection: pika.BlockingConnection, channel: pika.channel.Channel, method: str, *args: Any, **kwargs: Any
) -> None:
    if connection.is_open:
        cb = functools.partial(_channel_method_from_main, channel, method, *args, **kwargs)
        connection.add_callback_threadsafe(cb)
    else:
        logger.error("Can't %s as connection is closed or closing", method)


def _channel_method_from_main(channel: pika.channel.Channel, method: str, *args: Any, **kwargs: Any) -> None:
    if channel.is_open:
        getattr(channel, f"basic_{method}")(*args, **kwargs)
    else:
        logger.error("Can't %s as channel is closed or closing", method)



import pickle

import pytest

from yapw.util import default_encode, json_dumps


def test_json_dumps():
    assert json_dumps({"a": 1, "b": 2}) == b'{"a":1,"b":2}'


@pytest.mark.parametrize(
    "args,expected",
    [
        (({"a": 1, "b": 2}, "application/json"), b'{"a":1,"b":2}'),
        (("string", None), b"string"),
        ((b"bytes", None), b"bytes"),
        (({}, None), pickle.dumps({})),
    ],
)
def test_default_encode(args, expected):
    assert default_encode(*args) == expected



from unittest.mock import patch

from yapw.clients import Base, Blocking


class Client(Blocking, Base):
    pass


@patch("pika.BlockingConnection")
def test_init_default(connection):
    client = Client()

    connection.assert_called_once()

    assert connection.call_args[0][0].virtual_host == "/"
    assert connection.call_args[0][0].blocked_connection_timeout == 1800

    client.channel.basic_qos.assert_called_once_with(prefetch_count=1)


@patch("pika.BlockingConnection")
def test_init_kwargs(connection):
    client = Client(
        url="https://host:1234/%2Fv?blocked_connection_timeout=10", blocked_connection_timeout=300, prefetch_count=10
    )

    connection.assert_called_once()

    assert connection.call_args[0][0].virtual_host == "/v"
    assert connection.call_args[0][0].blocked_connection_timeout == 300

    client.channel.basic_qos.assert_called_once_with(prefetch_count=10)



#!/usr/bin/env python
"""
This script can be used to test signal handling.
"""

import logging
import time

from yapw import clients
from yapw.methods.blocking import ack

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


class Client(clients.Threaded, clients.Transient, clients.Blocking, clients.Base):
    pass


def callback(state, channel, method, properties, body):
    logger.info("Sleep")
    time.sleep(10)
    logger.info("Wake!")
    ack(state, channel, method.delivery_tag)


def main():
    client = Client(exchange="yapw_development")
    client.consume(callback, "sleep")


if __name__ == "__main__":
    main()



"""
Decorators to be used with consumer callbacks.

A message must be ack'd or nack'd if using `consumer prefetch <https://www.rabbitmq.com/consumer-prefetch.html>`__,
because otherwise `RabbitMQ stops delivering messages <https://www.rabbitmq.com/confirms.html#channel-qos-prefetch>`__.
The decorators help to ensure that, in case of error, either the message is nack'd or the process is halted.

:func:`~yapw.decorators.halt` is the default decorator. For example, if a callback inserts messages into a database,
and the database is down, but this exception isn't handled by the callback, then the :func:`~yapw.decorators.discard`
or :func:`~yapw.decorators.requeue` decorators would end up nack'ing all messages in the queue. The ``halt`` decorator
instead stops the consumer, so that an administrator can decide when it is appropriate to restart it.

Decorators look like this (see :func:`~yapw.decorators.decorate` for context):

.. code-block:: python

   from yapw.decorators import decorate


   def myfunction(decode, callback, state, channel, method, properties, body):
       def errback(exception):
           # do something, like halting the process or nack'ing the message

       decorate(decode, callback, state, channel, method, properties, body, errback)

User-defined decorators should avoid doing work outside the ``finally`` branch. Do work in the callback.
"""
from __future__ import annotations

import logging
import os
import signal
from typing import Any, Callable, Optional

import pika

from yapw.methods.blocking import nack
from yapw.types import ConsumerCallback, Decode, State
from yapw.util import jsonlib

logger = logging.getLogger(__name__)


def default_decode(body: bytes, content_type: Optional[str]) -> Any:
    """
    If the content type is "application/json", deserializes the JSON formatted bytes to a Python object. Otherwise,
    returns the bytes (which the consumer callback can deserialize independently).

    Uses `orjson <https://pypi.org/project/orjson/>`__ if available.

    :param body: the encoded message
    :param content_type: the message's content type
    :returns: a Python object
    """
    if content_type == "application/json":
        return jsonlib.loads(body)
    return body


def decorate(
    decode: Decode,
    callback: ConsumerCallback,
    state: State,
    channel: pika.channel.Channel,
    method: pika.spec.Basic.Deliver,
    properties: pika.BasicProperties,
    body: bytes,
    errback: Callable[[Exception], None],
    finalback: Optional[Callable[[], None]] = None,
) -> None:
    """
    Decode the message ``body`` using the ``decode`` function, and call the consumer ``callback``.

    If the ``callback`` function raises an exception, call the ``errback`` function. In any case, call the
    ``finalback`` function after calling the ``callback`` function.

    If the ``decode`` function raises an exception, send the SIGUSR2 signal to the main thread.

    .. seealso::

       :meth:`yapw.clients.Threaded.consume` for details on the consumer callback function signature.
    """
    logger.debug(
        "Received message %s with routing key %s and delivery tag %s", body, method.routing_key, method.delivery_tag
    )
    try:
        message = decode(body, properties.content_type)
        try:
            callback(state, channel, method, properties, message)
        except Exception as exception:
            errback(exception)
        finally:
            if finalback:
                finalback()
    except Exception:
        logger.exception("%r can't be decoded, sending SIGUSR2", body)
        os.kill(os.getpid(), signal.SIGUSR2)


# https://stackoverflow.com/a/7099229/244258
def halt(
    decode: Decode,
    callback: ConsumerCallback,
    state: State,
    channel: pika.channel.Channel,
    method: pika.spec.Basic.Deliver,
    properties: pika.BasicProperties,
    body: bytes,
) -> None:
    """
    If the callback raises an exception, send the SIGUSR1 signal to the main thread, without acknowledgment.
    """

    def errback(exception: Exception) -> None:
        logger.exception("Unhandled exception when consuming %r, sending SIGUSR1", body)
        os.kill(os.getpid(), signal.SIGUSR1)

    decorate(decode, callback, state, channel, method, properties, body, errback)


def discard(
    decode: Decode,
    callback: ConsumerCallback,
    state: State,
    channel: pika.channel.Channel,
    method: pika.spec.Basic.Deliver,
    properties: pika.BasicProperties,
    body: bytes,
) -> None:
    """
    If the callback raises an exception, nack the message without requeueing.
    """

    def errback(exception: Exception) -> None:
        logger.exception("Unhandled exception when consuming %r, discarding message", body)
        nack(state, channel, method.delivery_tag, requeue=False)

    decorate(decode, callback, state, channel, method, properties, body, errback)


def requeue(
    decode: Decode,
    callback: ConsumerCallback,
    state: State,
    channel: pika.channel.Channel,
    method: pika.spec.Basic.Deliver,
    properties: pika.BasicProperties,
    body: bytes,
) -> None:
    """
    If the callback raises an exception, nack the message, and requeue the message unless it was redelivered.
    """

    def errback(exception: Exception) -> None:
        requeue = not method.redelivered
        logger.exception("Unhandled exception when consuming %r (requeue=%r)", body, requeue)
        nack(state, channel, method.delivery_tag, requeue=requeue)

    decorate(decode, callback, state, channel, method, properties, body, errback)



from setuptools import find_packages, setup

with open("README.rst") as f:
    long_description = f.read()

setup(
    name="yapw",
    version="0.0.13",
    author="Open Contracting Partnership",
    author_email="data@open-contracting.org",
    url="https://github.com/open-contracting/yapw",
    description="A Pika wrapper with error handling, signal handling and good defaults.",
    license="BSD",
    packages=find_packages(exclude=["tests", "tests.*"]),
    long_description=long_description,
    long_description_content_type="text/x-rst",
    install_requires=[
        "pika>=1.2.0",
        "typing-extensions;python_version<'3.8'",
    ],
    extras_require={
        "perf": [
            "orjson",
        ],
        "test": [
            "coveralls",
            "pytest",
            "pytest-cov",
        ],
        "types": [
            "mypy",
            "types-orjson",
            "types-pika",
        ],
        "docs": [
            "furo",
            "sphinx",
            "sphinx-autobuild",
        ],
    },
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
)
